import zookeeper

from .named_value import NamedValue


class Event(NamedValue):
  MAP = {
    0: 'UNKNOWN',
    zookeeper.CREATED_EVENT: 'CREATED',
    zookeeper.DELETED_EVENT: 'DELETED',
    zookeeper.CHANGED_EVENT: 'CHANGED',
    zookeeper.CHILD_EVENT: 'CHILD',
    zookeeper.SESSION_EVENT: 'SESSION',
    zookeeper.NOTWATCHING_EVENT: 'NOTWATCHING'
  }

  @property
  def map(self):
    return self.MAP


class State(NamedValue):
  MAP = {
    0: 'UNKNOWN',
    zookeeper.CONNECTING_STATE: 'CONNECTING',
    zookeeper.ASSOCIATING_STATE: 'ASSOCIATING',
    zookeeper.CONNECTED_STATE: 'CONNECTED',
    zookeeper.EXPIRED_SESSION_STATE: 'EXPIRED_SESSION',
    zookeeper.AUTH_FAILED_STATE: 'AUTH_FAILED',
  }

  @property
  def map(self):
    return self.MAP


class ReturnCode(NamedValue):
  MAP = {
    # Normal
    zookeeper.OK: 'OK',

    # Abnormal
    zookeeper.NONODE: 'NONODE',
    zookeeper.NOAUTH: 'NOAUTH',
    zookeeper.BADVERSION: 'BADVERSION',
    zookeeper.NOCHILDRENFOREPHEMERALS: 'NOCHILDRENFOREPHEMERALS',
    zookeeper.NODEEXISTS: 'NODEEXISTS',
    zookeeper.NOTEMPTY: 'NOTEMPTY',
    zookeeper.SESSIONEXPIRED: 'SESSIONEXPIRED',
    zookeeper.INVALIDCALLBACK: 'INVALIDCALLBACK',
    zookeeper.INVALIDACL: 'INVALIDACL',
    zookeeper.AUTHFAILED: 'AUTHFAILED',
    zookeeper.CLOSING: 'CLOSING',
    zookeeper.NOTHING: 'NOTHING',
    zookeeper.SESSIONMOVED: 'SESSIONMOVED',

    # Exceptional
    zookeeper.SYSTEMERROR: 'SYSTEMERROR',
    zookeeper.RUNTIMEINCONSISTENCY: 'RUNTIMEINCONSISTENCY',
    zookeeper.DATAINCONSISTENCY: 'DATAINCONSISTENCY',
    zookeeper.CONNECTIONLOSS: 'CONNECTIONLOSS',
    zookeeper.MARSHALLINGERROR: 'MARSHALLINGERROR',
    zookeeper.UNIMPLEMENTED: 'UNIMPLEMENTED',
    zookeeper.OPERATIONTIMEOUT: 'OPERATIONTIMEOUT',
    zookeeper.BADARGUMENTS: 'BADARGUMENTS',
    zookeeper.INVALIDSTATE: 'INVALIDSTATE'
  }

  @property
  def map(self):
    return self.MAP


class Id(object):
  def __init__(self, scheme, id_):
    if not isinstance(scheme, str) or not isinstance(id_, str):
      raise ValueError('Scheme and id must be strings!')
    self.scheme = scheme
    self.id = id_


class Acl(dict):
  def __init__(self, perm, id_):
    dict.__init__(self)
    self['perms'] = perm
    self['scheme'] = id_.scheme
    self['id'] = id_.id



import logging
import threading
import sys

from twitter.common.metrics import (
    AtomicGauge,
    LambdaGauge,
    Observable)

from kazoo.client import KazooClient
from kazoo.protocol.states import KazooState, KeeperState
from kazoo.retry import KazooRetry


DEFAULT_RETRY_MAX_DELAY_SECS = 600


DEFAULT_RETRY_DICT = dict(
    max_tries=None,
    ignore_expire=True,
)


class TwitterKazooClient(KazooClient, Observable):
  @classmethod
  def make(cls, *args, **kw):
    # TODO(jcohen): Consider removing verbose option entirely in favor of just using loglevel.
    verbose = kw.pop('verbose', False)
    async = kw.pop('async', True)

    if verbose:
      loglevel = kw.pop('loglevel', logging.INFO)
    else:
      loglevel = kw.pop('loglevel', sys.maxsize)

    logger = logging.getLogger('kazoo.devnull')
    logger.setLevel(loglevel)
    kw['logger'] = logger

    zk = cls(*args, **kw)
    if async:
      zk.start_async()
      zk.connecting.wait()
    else:
      zk.start()

    return zk

  def __init__(self, *args, **kw):
    if 'connection_retry' not in kw:
      # The default backoff delay limit in kazoo is 3600 seconds, which is generally
      # too conservative for our use cases.  If not supplied by the caller, provide
      # a backoff that will truncate earlier.
      kw['connection_retry'] = KazooRetry(
          max_delay=DEFAULT_RETRY_MAX_DELAY_SECS, **DEFAULT_RETRY_DICT)

    super(TwitterKazooClient, self).__init__(*args, **kw)
    self.connecting = threading.Event()
    self.__session_expirations = AtomicGauge('session_expirations')
    self.__connection_losses = AtomicGauge('connection_losses')
    self.__session_id = LambdaGauge('session_id', lambda: (self._session_id or 0))
    self.metrics.register(self.__session_expirations)
    self.metrics.register(self.__connection_losses)
    self.metrics.register(self.__session_id)
    self.add_listener(self._observable_listener)

  def _observable_listener(self, state):
    if state == KazooState.LOST:
      self.__session_expirations.increment()
    elif state == KazooState.SUSPENDED:
      self.__connection_losses.increment()

  def _session_callback(self, state):
    rc = super(TwitterKazooClient, self)._session_callback(state)
    if state == KeeperState.CONNECTING:
      self.connecting.set()
    return rc

  @property
  def live(self):
    return self._live



import json
import socket
from thrift.TSerialization import deserialize as thrift_deserialize

from gen.twitter.thrift.endpoint.ttypes import (
  Endpoint as ThriftEndpoint,
  ServiceInstance as ThriftServiceInstance)

try:
  from twitter.common import log
except ImportError:
  import logging as log

from twitter.common.lang import Compatibility


class Endpoint(object):
  @classmethod
  def unpack_thrift(cls, blob):
    return cls(blob.host, blob.port, blob.inet, blob.inet6)

  @classmethod
  def from_dict(cls, value):
    inet = value.get('inet')
    inet6 = value.get('inet6')
    return Endpoint(value['host'], value['port'], inet, inet6)

  @classmethod
  def to_dict(cls, endpoint):
    d = {
      'host': endpoint.host,
      'port': endpoint.port
    }
    if endpoint.inet is not None:
      d['inet'] = endpoint.inet
    if endpoint.inet6 is not None:
      d['inet6'] = endpoint.inet6
    return d

  def __init__(self, host, port, inet=None, inet6=None):
    """host -- the hostname of the device where this endpoint can be found
       port -- the port number the device serves this endpoint on
       inet -- a human-readable representation of the  IPv4 address of the device
               where this endpoint can be found.
       inet6 -- a human-readable representation of the IPv6 address of the device
                where this endpoint can be found.

    `inet` and/or `inet6` can be used when present to avoid a DNS lookup.

    """
    if not isinstance(host, Compatibility.string):
      raise ValueError('Expected host to be a string!')
    if not isinstance(port, int):
      raise ValueError('Expected port to be an integer!')
    if inet is not None:
      try:
        socket.inet_pton(socket.AF_INET, inet)
      except socket.error:
        raise ValueError('Expected "%s" to be a string containing a valid IPv4 address!' % inet)
    if inet6 is not None:
      try:
        socket.inet_pton(socket.AF_INET6, inet6)
      except socket.error:
        raise ValueError('Expected "%s" to be a string containing a valid IPv6 address!' % inet6)

    self._host = host
    self._port = port
    self._inet = inet
    self._inet6 = inet6

  def __key(self):
    return (self.host, self.port, self._inet, self._inet6)

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__key() == other.__key()

  def __hash__(self):
    return hash(self.__key())

  @property
  def host(self):
    return self._host

  @property
  def port(self):
    return self._port

  @property
  def inet(self):
    return self._inet

  @property
  def inet6(self):
    return self._inet6

  def __str__(self):
    return '%s:%s' % (self.host, self.port)


class Status(object):
  MAP = {
    0: 'DEAD',
    1: 'STARTING',
    2: 'ALIVE',
    3: 'STOPPING',
    4: 'STOPPED',
    5: 'WARNING'
  }

  @staticmethod
  def from_id(_id):
    if _id not in Status.MAP:
      raise ValueError('from_id got an invalid Status!')
    return Status(Status.MAP.get(_id), _id)

  @staticmethod
  def from_string(string):
    for _id, name in Status.MAP.items():
      if name == string:
        return Status(name, _id)
    raise ValueError('from_string got an invalid Status!')

  from_thrift = from_id

  def __init__(self, name, _id):
    self._name = name
    self._id = _id

  def name(self):
    return self._name

  def __eq__(self, other):
    return self._id == other._id

  def __hash__(self):
    return hash(self._id)

  def __str__(self):
    return self.name()


class ServiceInstance(object):
  class InvalidType(Exception): pass
  class UnknownEndpoint(Exception): pass

  @classmethod
  def unpack(cls, blob, member_id=None):
    """Return a ServiceInstance object from the json or thrift implementation of
      a serverset member.
      The member_id needs to be passed in separately because the info method from the group
      (either Group or KazooGroup) does not include it.
    """
    try:
      return cls.unpack_json(blob, member_id)
    except Exception as e1:
      try:
        return cls.unpack_thrift(blob, member_id)
      except Exception as e2:
        log.debug('Failed to deserialize JSON: %s (%s) && Thrift: %s (%s)' % (
          e1, e1.__class__.__name__, e2, e2.__class__.__name__))
        return None

  @classmethod
  def unpack_json(cls, blob, member_id=None):
    blob = json.loads(blob)
    for key in ('status', 'serviceEndpoint', 'additionalEndpoints'):
      if key not in blob:
        raise ValueError('Expected to find %s in ServiceInstance JSON!' % key)
    additional_endpoints = dict((name, Endpoint.from_dict(value))
      for name, value in blob['additionalEndpoints'].items())
    shard = cls.__check_int(blob.get('shard'))
    member_id = cls.__check_int(member_id)
    return cls(
      service_endpoint=Endpoint.from_dict(blob['serviceEndpoint']),
      additional_endpoints=additional_endpoints,
      status=Status.from_string(blob['status']),
      shard=shard,
      member_id=member_id)

  @classmethod
  def unpack_thrift(cls, blob, member_id=None):
    if not isinstance(blob, ThriftServiceInstance):
      blob = thrift_deserialize(ThriftServiceInstance(), blob)
    additional_endpoints = dict((name, Endpoint.unpack_thrift(value))
      for name, value in blob.additionalEndpoints.items())
    member_id = cls.__check_int(member_id)
    return cls(
      service_endpoint=Endpoint.unpack_thrift(blob.serviceEndpoint),
      additional_endpoints=additional_endpoints,
      status=Status.from_thrift(blob.status),
      shard=blob.shard,
      member_id=member_id)

  @classmethod
  def to_dict(cls, service_instance):
    instance = dict(
      serviceEndpoint=Endpoint.to_dict(service_instance.service_endpoint),
      additionalEndpoints=dict((name, Endpoint.to_dict(endpoint))
          for name, endpoint in service_instance.additional_endpoints.items()),
      status=service_instance.status.name()
    )
    if service_instance.shard is not None:
      instance.update(shard=service_instance.shard)
    if service_instance.member_id is not None:
      instance.update(member_id=service_instance.member_id)
    return instance

  @classmethod
  def pack(cls, service_instance):
    return json.dumps(cls.to_dict(service_instance))

  def __init__(
      self,
      service_endpoint,
      additional_endpoints=None,
      status='ALIVE',
      shard=None,
      member_id=None):
    if not isinstance(service_endpoint, Endpoint):
      raise ValueError('Expected service_endpoint to be an Endpoint, got %r' % service_endpoint)
    self._shard = shard
    self._member_id = member_id
    self._service_endpoint = service_endpoint
    self._additional_endpoints = additional_endpoints or {}
    if not isinstance(self._additional_endpoints, dict):
      raise ValueError('Additional endpoints must be a dictionary.')
    for name, endpoint in self._additional_endpoints.items():
      if not isinstance(name, Compatibility.string):
        raise ValueError('Expected additional endpoints to be named by strings.')
      if not isinstance(endpoint, Endpoint):
        raise ValueError('Endpoints must be of type Endpoint.')
    if isinstance(status, Compatibility.string):
      self._status = Status.from_string(status)
      if self._status is None:
        raise ValueError('Unknown status: %s' % status)
    elif isinstance(status, Status):
      self._status = status
    else:
      raise ValueError('Status must be of type ServiceInstance.Status or string.')

  @property
  def service_endpoint(self):
    return self._service_endpoint

  @property
  def additional_endpoints(self):
    return self._additional_endpoints

  @property
  def status(self):
    return self._status

  @property
  def shard(self):
    return self._shard

  @property
  def member_id(self):
    return self._member_id

  @staticmethod
  def __check_int(item):
    if item is not None:
      try:
        item = int(item)
      except ValueError:
        log.warn('Failed to deserialize value %r' % item)
        item = None
    return item

  def __additional_endpoints_string(self):
    return ['%s=>%s' % (key, val) for key, val in self.additional_endpoints.items()]

  def __key(self):
    return (
        self.service_endpoint,
        frozenset(sorted(self.__additional_endpoints_string())),
        self.status,
        self._shard,
        self._member_id)

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__key() == other.__key()

  def __hash__(self):
    return hash(self.__key())

  def __str__(self):
    return 'ServiceInstance(%s, %s, %saddl: %s, status: %s)' % (
      self.service_endpoint,
      ('shard: %s, ' % self._shard) if self._shard is not None else '',
      ('member_id: %s, ' % self._member_id) if self._member_id is not None else '',
      ' : '.join(self.__additional_endpoints_string()),
      self.status)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from collections import namedtuple
from functools import wraps
import posixpath
import random
import socket
import sys
import threading
import zookeeper

try:
  from twitter.common import app
  WITH_APP=True
except ImportError:
  WITH_APP=False

try:
  from twitter.common import log
  from twitter.common.log.options import LogOptions
except ImportError:
  import logging as log

from twitter.common.metrics import (
    AtomicGauge,
    LambdaGauge,
    Observable)

try:
  from Queue import Queue, Empty
except ImportError:
  from queue import Queue, Empty

from .constants import Acl, Id


if WITH_APP:
  app.add_option(
      '--zookeeper',
      default='zookeeper.local.twitter.com:2181',
      metavar='HOST:PORT[,HOST:PORT,...]',
      dest='twitter_common_zookeeper_ensemble',
      help='A comma-separated list of host:port of ZooKeeper servers.')
  app.add_option(
      '--zookeeper_timeout',
      type='float',
      default=15.0,
      dest='twitter_common_zookeeper_timeout',
      help='The default timeout (in seconds) for ZK operations.')
  app.add_option(
      '--zookeeper_reconnects',
      type='int',
      default=0,
      dest='twitter_common_zookeeper_reconnects',
      help='The number of permitted reconnections before failing zookeeper (0 = infinite).')
  app.add_option(
      '--zookeeper_log_level',
      dest='twitter_common_zookeeper_log_level_override',
      choices=('NONE', 'DEBUG','INFO','WARN','ERROR','FATAL'),
      default='',
      help='Override the default ZK logging level.')

  class ZookeeperLoggingSubsystem(app.Module):
    # Map the ZK debug log to the same level as stderr logging.
    _ZK_LOG_LEVEL_MAP = {
        log.DEBUG: zookeeper.LOG_LEVEL_DEBUG,
        log.INFO: zookeeper.LOG_LEVEL_INFO,
        log.WARN: zookeeper.LOG_LEVEL_WARN,
        log.ERROR: zookeeper.LOG_LEVEL_ERROR,
        log.FATAL: zookeeper.LOG_LEVEL_ERROR,
    }

    def __init__(self):
      app.Module.__init__(self, __name__, description='Zookeeper logging subsystem.')

    def setup_function(self):
      log_level_override = app.get_options().twitter_common_zookeeper_log_level_override
      self._set_log_level(log_level_override=log_level_override)

    def _set_log_level(self, log_level_override=''):
      stderr_log_level = LogOptions.stderr_log_level()
      # set default level to FATAL.
      # we do this here (instead of add_option) to distinguish when an override is set.
      if stderr_log_level == log.INFO and log_level_override != 'INFO':
        stderr_log_level = log.FATAL
      # default to using stderr logging level, setting override if applicable
      log_level = getattr(log, log_level_override, stderr_log_level)
      # set the logger
      zk_log_level = ZookeeperLoggingSubsystem._ZK_LOG_LEVEL_MAP.get(
          log_level, zookeeper.LOG_LEVEL_ERROR)
      zookeeper.set_debug_level(zk_log_level)


  app.register_module(ZookeeperLoggingSubsystem())


class Perms(object):
  READ = zookeeper.PERM_READ
  WRITE = zookeeper.PERM_WRITE
  CREATE = zookeeper.PERM_CREATE
  DELETE = zookeeper.PERM_DELETE
  ADMIN = zookeeper.PERM_ADMIN
  ALL = zookeeper.PERM_ALL


class Ids(object):
  ANYONE_ID_UNSAFE = Id('world', 'anyone')
  AUTH_IDS = Id('auth', '')


class Acls(object):
  OPEN_ACL_UNSAFE = [Acl(Perms.ALL, Ids.ANYONE_ID_UNSAFE)]
  CREATOR_ALL_ACL = [Acl(Perms.ALL, Ids.AUTH_IDS)]
  READ_ACL_UNSAFE = [Acl(Perms.READ, Ids.ANYONE_ID_UNSAFE)]
  EVERYONE_READ_CREATOR_ALL = [Acl(Perms.ALL, Ids.AUTH_IDS), Acl(Perms.READ, Ids.ANYONE_ID_UNSAFE)]


class ZooDefs(object):
  Acls = Acls
  Ids = Ids
  Perms = Perms


del Acls, Ids, Perms


class ZooKeeper(Observable):
  """A convenience wrapper around the low-level ZooKeeper API.

  Blocks until the initial connection is established, and proxies method calls
  to the corresponding ZK functions, passing the handle.

  Supports both synchronous and asynchronous APIs.

  Syncronous API Notes:

    Synchronous calls will block across connection loss or session
    expiration until reconnected.

  Asynchronous API Notes:

    Asynchronous calls will queue up while the session/connection is
    unhealthy and only be dispatched while zookeeper is healthy.  It is
    still possible for asynchronous calls to fail should the session be
    severed after the call has been successfully dispatched.  In other
    words: don't assume your rc will always be zookeeper.OK.

    Watches will behave as normal assuming successful dispatch.  In general
    when using this wrapper, you should retry your call if your watch is
    fired with EXPIRED_SESSION_STATE and ignore anything else whose state is
    not CONNECTED_STATE.  This wrapper will never re-dispatch calls that
    have been sent to zookeeper without error.
  """

  class Error(Exception): pass
  class ConnectionTimeout(Error): pass
  class InvalidEnsemble(Error): pass
  class Stopped(Error): pass

  # White-list of methods that accept a ZK handle as their first argument
  _ZK_SYNC_METHODS = frozenset([
      'add_auth', 'close', 'create', 'delete', 'exists', 'get', 'get_acl',
      'get_children', 'is_unrecoverable', 'recv_timeout', 'set', 'set2',
      'set_acl', 'set_watcher', 'state',
   ])

  _ZK_ASYNC_METHODS = frozenset([
      'acreate', 'adelete', 'aexists', 'aget', 'aget_acl', 'aget_children', 'aset',
      'aset_acl', 'async'
  ])

  COMPLETION_RETRY = frozenset([
    zookeeper.CONNECTIONLOSS,
    zookeeper.OPERATIONTIMEOUT,
    zookeeper.SESSIONEXPIRED,
    zookeeper.CLOSING,
  ])

  @classmethod
  def expand_ensemble(cls, servers):
    """Expand comma-separated list of host:port to comma-separated, fully-resolved list of ip:port."""
    server_ports = []
    for server_port in servers.split(','):
      server_split = server_port.split(':', 2)
      if len(server_split) == 1:
        server, port = server_split[0], cls.DEFAULT_PORT
      else:
        try:
          server, port = server_split[0], int(server_split[1])
        except ValueError:
          raise cls.InvalidEnsemble('Invalid ensemble string: %s' % server_port)
      try:
        for ip in socket.gethostbyname_ex(server)[2]:
          server_ports.append('%s:%s' % (ip, port))
      except socket.gaierror:
        raise cls.InvalidEnsemble('Could not resolve %s' % server)
    return ','.join(server_ports)

  DEFAULT_TIMEOUT_SECONDS = 30.0
  DEFAULT_ENSEMBLE = 'localhost:2181'
  DEFAULT_PORT = 2181
  DEFAULT_ACL = ZooDefs.Acls.OPEN_ACL_UNSAFE
  MAX_RECONNECTS = 1

  # (is live?, is stopped?) => human readable status
  STATUS_MATRIX = {
    (True, True): 'WTF',
    (True, False): 'OK',
    (False, True): 'STOPPED',
    (False, False): 'CONNECTING'
  }

  class Completion(object):
    def __init__(self, zk, function, *args, **kw):
      self._zk = zk
      self._cid = random.randint(0, sys.maxint - 1)
      self._logger = kw.pop('logger', log.debug)
      @wraps(function)
      def wrapper(zh):
        return function(zh, *args, **kw)
      self._fn = wrapper
      self._logger('Created %s args:(%s) kw:{%s}' % (
        self,
        ', '.join(map(repr, args)),
        ', '.join('%s: %r' % (key, val) for key, val in kw.items())))

    def __str__(self):
      return '%s(id:%s, zh:%s, %s)' % (
          self.__class__.__name__, self._cid, self._zk._zh, self._fn.__name__)

    def __call__(self):
      try:
        self._logger('%s start' % self)
        result = self._fn(self._zk._zh)
        self._logger('%s success' % self)
        return result
      except TypeError as e:
        # Raced; zh now dead, so re-enqueue.
        if self._zk._zh is not None:
          raise
        self._logger('%s raced, re-enqueueing' % self)
        self._zk._add_completion(self._fn)
      except (zookeeper.ConnectionLossException,
              zookeeper.InvalidStateException,
              zookeeper.SessionExpiredException,
              SystemError) as e:
        self._logger('%s excepted (%s), re-enqueueing' % (self, e))
        self._zk._add_completion(self._fn)
      return zookeeper.OK

  # N.B.(wickman) This is code is theoretically racy.  We cannot synchronize
  # events across the zookeeper C event loop, however we do everything in
  # our power to catch transitional latches.  These are almost always
  # exercised in tests and never in practice.
  #
  # TODO(wickman) ConnectionLoss probably does not encapsulate all the
  # exception states that arise on connection loss and/or session
  # expiration.  However, we don't want to blanket catch ZooKeeperException
  # because some things e.g.  get() will raise NoNodeException.  We should
  # partition the exception space in two: behavioral exceptions and, well,
  # exceptional exceptions.
  class BlockingCompletion(Completion):
    def __call__(self):
      while True:
        try:
          self._logger('%s start' % self)
          result = self._fn(self._zk._zh)
          self._logger('%s success' % self)
          return result
        except (zookeeper.ConnectionLossException,
                zookeeper.InvalidStateException,
                zookeeper.SessionExpiredException,
                TypeError) as e:
          # TypeError because we raced on live latch from True=>False when _zh gets reinitialized.
          if isinstance(e, TypeError) and self._zk._zh is not None:
            self._logger('%s excepted, user error' % self)
            raise
          # We had the misfortune of the live latch being set but having a session event propagate
          # before the BlockingCompletion could be executed.
          while not self._zk._stopped.is_set():
            self._logger('%s [live: %s] excepted on connection event: %s' % (
                self, self._zk._live.is_set(), e))
            self._zk._live.wait(timeout=0.1)
            if self._zk._live.is_set():
              break
          if self._zk._stopped.is_set():
            raise ZooKeeper.Stopped('ZooKeeper is stopped.')
        except Exception as e:
          self._logger('%s excepted unexpectedly: %s' % (self, e))
          raise

  def __init__(self,
               servers=None,
               timeout_secs=None,
               watch=None,
               max_reconnects=None,
               authentication=None,
               logger=log.debug):
    """Create new ZooKeeper object.

    Blocks until ZK negotation completes, or the timeout expires. By default
    only tries to connect once.  Use a larger 'max_reconnects' if you want
    to be resilient to things such as DNS outages/changes.

    If watch is set to a function, it is called whenever the global
    zookeeper watch is dispatched using the same function signature, with the
    exception that this object is used in place of the zookeeper handle.

    If authentication is set, it should be a tuple of (scheme, credentials),
    for example, ('digest', 'username:password')
    """

    default_ensemble = self.DEFAULT_ENSEMBLE
    default_timeout = self.DEFAULT_TIMEOUT_SECONDS
    default_reconnects = self.MAX_RECONNECTS
    if WITH_APP:
      options = app.get_options()
      default_ensemble = options.twitter_common_zookeeper_ensemble
      default_timeout = options.twitter_common_zookeeper_timeout
      default_reconnects = options.twitter_common_zookeeper_reconnects
    self._servers = servers or default_ensemble
    self._timeout_secs = timeout_secs or default_timeout
    self._init_count = 0
    self._credentials = authentication
    self._authenticated = threading.Event()
    self._live = threading.Event()
    self._stopped = threading.Event()
    self._completions = Queue()
    self._zh = None
    self._watch = watch
    self._logger = logger
    self._max_reconnects = max_reconnects if max_reconnects is not None else default_reconnects
    self._init_metrics()
    self.reconnect()

  def __del__(self):
    self._safe_close()

  def _log(self, msg):
    self._logger('[zh:%s] %s' % (self._zh, msg))

  def _init_metrics(self):
    self._session_expirations = AtomicGauge('session_expirations')
    self._connection_losses = AtomicGauge('connection_losses')
    self.metrics.register(self._session_expirations)
    self.metrics.register(self._connection_losses)
    self.metrics.register(LambdaGauge('session_id', lambda: self.session_id))
    self.metrics.register(LambdaGauge('live', lambda: int(self._live.is_set())))

  @property
  def session_id(self):
    try:
      session_id, _ = zookeeper.client_id(self._zh)
      return session_id
    except:
      return None

  @property
  def session_expirations(self):
    return self._session_expirations.read()

  @property
  def connection_losses(self):
    return self._connection_losses.read()

  @property
  def live(self):
    return self._live

  def stop(self):
    """Gracefully stop this Zookeeper session."""
    self._log('Shutting down ZooKeeper')
    self._stopped.set()
    self._safe_close()
    self._completions = Queue()  # there is no .clear()

  def restart(self):
    """Stop and restart this Zookeeper session.  Unfinished completions will be retried
       on reconnection."""
    self._safe_close()
    self._stopped.clear()
    self.reconnect()

  def _safe_close(self):
    if self._zh is not None:
      zh, self._zh = self._zh, None
      try:
        zookeeper.close(zh)
      except zookeeper.ZooKeeperException:
        # the session has been corrupted or otherwise disconnected
        pass
      self._live.clear()

  def _add_completion(self, function, *args, **kw):
    self._completions.put(self.Completion(self, function, logger=self._log, *args, **kw))

  def _clear_completions(self):
    while self._live.is_set():
      try:
        completion = self._completions.get_nowait()
        completion()
        self._completions.task_done()
      except Empty:
        return

  def reconnect(self):
    """Attempt to reconnect to ZK."""
    if self._stopped.is_set():
      self._safe_close()
      return

    def safe_close(zh):
      try:
        zookeeper.close(zh)
      except:
        # TODO(wickman) When the SystemError bug is fixed in zkpython, narrow this except clause.
        pass

    def activate():
      self._authenticated.set()
      self._live.set()

    def on_authentication(zh, rc):
      if self._zh != zh:
        safe_close(zh)
        return
      if rc == zookeeper.OK:
        activate()

    def maybe_authenticate():
      if self._authenticated.is_set() or not self._credentials:
        activate()
        return
      try:
        scheme, credentials = self._credentials
        zookeeper.add_auth(self._zh, scheme, credentials, on_authentication)
      except zookeeper.ZooKeeperException as e:
        self._logger('Failed to authenticate: %s' % e)

    def connection_handler(handle, type, state, path):
      if self._zh != handle:
        safe_close(handle)
        return
      if self._stopped.is_set():
        return
      if self._watch:
        self._watch(self, type, state, path)
      if state == zookeeper.CONNECTED_STATE:
        self._logger('Connection started, setting live.')
        maybe_authenticate()
        self._clear_completions()
      elif state == zookeeper.EXPIRED_SESSION_STATE:
        self._logger('Session lost, clearing live state.')
        self._session_expirations.increment()
        self._live.clear()
        self._authenticated.clear()
        self._zh = None
        self._init_count = 0
        self.reconnect()
      else:
        self._logger('Connection lost, clearing live state.')
        self._connection_losses.increment()
        self._live.clear()

    # this closure is exposed for testing only -- in order to simulate session events.
    self._handler = connection_handler

    timeout_ms = int(self._timeout_secs * 1000)
    while True:
      self._safe_close()
      servers = self.expand_ensemble(self._servers)
      self._log('Connecting to ZK hosts at %s' % servers)
      self._zh = zookeeper.init(servers, connection_handler, timeout_ms)
      self._init_count += 1
      self._live.wait(self._timeout_secs + 1)
      if self._live.is_set():
        break
      elif self._max_reconnects > 0 and self._init_count >= self._max_reconnects:
        self._safe_close()
        raise ZooKeeper.ConnectionTimeout('Timed out waiting for ZK connection to %s' % servers)
    self._log('Successfully connected to ZK at %s' % servers)

  def _wrap_sync(self, function_name):
    """Wrap a zookeeper module function in an error-handling completion that injects the
       current zookeeper handle as the first parameter."""
    function = getattr(zookeeper, function_name)
    @wraps(function)
    def _curry(*args, **kwargs):
      return self.BlockingCompletion(self, function, logger=self._log, *args, **kwargs)()
    return _curry

  def _wrap_async(self, function_name):
    """Wrap an asynchronous zookeeper module function in an error-handling
       completion that injects the current zookeeper handle as the first
       parameter and puts it on a completion queue if the current connection
       state is unhealthy."""
    function = getattr(zookeeper, function_name)
    @wraps(function)
    def _curry(*args, **kwargs):
      completion = self.Completion(self, function, logger=self._log, *args, **kwargs)
      if self._live.is_set():
        return completion()
      else:
        # TODO(wickman)  This is racy, should it go from not live => live
        # prior to Queue.put.  Two solutions: a periodic background thread
        # that attempts to empty the completion queue, or use a mutex-protected
        # container for self._live.
        self._completions.put(self.Completion(self, function, logger=self._log, *args, **kwargs))
        return zookeeper.OK  # proxy OK.
    return _curry

  def safe_create(self, path, acl=DEFAULT_ACL):
    child = '/'
    for component in filter(None, path.split('/')):
      child = posixpath.join(child, component)
      try:
        self.create(child, "", acl, 0)
      except zookeeper.NodeExistsException:
        continue
      except zookeeper.NoAuthException:
        if not self.exists(child):
          raise
    return child

  def safe_delete(self, path):
    try:
      if not self.exists(path):
        return True
      for child in self.get_children(path):
        if not self.safe_delete(posixpath.join(path, child)):
          return False
      self.delete(path)
    except zookeeper.ZooKeeperException:
      return False
    return True

  def __getattr__(self, function_name):
    """Proxy to underlying ZK functions."""
    if function_name in ZooKeeper._ZK_SYNC_METHODS:
      return self._wrap_sync(function_name)
    elif function_name in ZooKeeper._ZK_ASYNC_METHODS:
      return self._wrap_async(function_name)
    else:
      raise AttributeError('%r has no attribute %r' % (self, function_name))

  def __str__(self):
    return 'ZooKeeper(status=%s,queued=%d,servers=%r)' % (
      self.STATUS_MATRIX[(self._live.is_set(), self._stopped.is_set())],
      self._completions.qsize(), self._servers)

  def __repr__(self):
    return 'ZooKeeper(servers=%r)' % self._servers



from twitter.common import app

from kazoo.client import KazooClient
from twitter.common.zookeeper.group.kazoo_group import KazooGroup


def main():
  import code
  code.interact(local=globals())


app.main()



__all__ = (
  'Endpoint',
  'ServerSet',
  'ServiceInstance',
  'get_serverset_hosts',
)

from .endpoint import Endpoint, ServiceInstance
from .serverset import ServerSet


def get_serverset_hosts(serverset_path, zk):
  ss = ServerSet(zk, serverset_path)
  return list(ss)



import functools
import posixpath
import threading
import time

try:
  from twitter.common import log
except ImportError:
  import logging as log

from twitter.common.concurrent import Future
from twitter.common.exceptions import ExceptionalThread
from twitter.common.quantity import Amount, Time
from twitter.common.zookeeper.constants import ReturnCode

from .group_base import (
    Capture,
    GroupBase,
    GroupInterface,
    Membership,
    set_different)

import zookeeper


class Group(GroupBase, GroupInterface):
  """
    An implementation of GroupInterface against CZookeeper.
  """

  def __init__(self, zk, path, acl=None):
    self._zk = zk
    self._path = '/' + '/'.join(filter(None, path.split('/')))  # normalize path
    self._members = {}
    self._member_lock = threading.Lock()
    self._acl = acl or zk.DEFAULT_ACL

  def _prepare_path(self, success):
    class Background(ExceptionalThread):
      BACKOFF = Amount(5, Time.SECONDS)
      def run(_):
        child = '/'
        for component in self._path.split('/')[1:]:
          child = posixpath.join(child, component)
          while True:
            try:
              self._zk.create(child, "", self._acl)
              break
            except zookeeper.NodeExistsException:
              break
            except zookeeper.NoAuthException:
              if self._zk.exists(child):
                break
              else:
                success.set(False)
                return
            except zookeeper.OperationTimeoutException:
              time.sleep(Background.BACKOFF.as_(Time.SECONDS))
              continue
        success.set(True)
    background = Background()
    background.daemon = True
    background.start()

  def info(self, member, callback=None):
    if member == Membership.error():
      raise self.InvalidMemberError('Cannot get info on error member!')

    capture = Capture(callback)

    def do_info():
      self._zk.aget(path, None, info_completion)

    with self._member_lock:
      if member not in self._members:
        self._members[member] = Future()
      member_future = self._members[member]

    member_future.add_done_callback(lambda future: capture.set(future.result()))

    dispatch = False
    with self._member_lock:
      if not member_future.done() and not member_future.running():
        try:
          dispatch = member_future.set_running_or_notify_cancel()
        except:
          pass

    def info_completion(_, rc, content, stat):
      if rc in self._zk.COMPLETION_RETRY:
        do_info()
        return
      if rc == zookeeper.NONODE:
        future = self._members.pop(member, Future())
        future.set_result(Membership.error())
        return
      elif rc != zookeeper.OK:
        return
      self._members[member].set_result(content)

    if dispatch:
      path = posixpath.join(self._path, self.id_to_znode(member.id))
      do_info()

    return capture()

  def join(self, blob, callback=None, expire_callback=None):
    membership_capture = Capture(callback)
    exists_capture = Capture(expire_callback)

    def on_prepared(success):
      if success:
        do_join()
      else:
        membership_capture.set(Membership.error())

    prepare_capture = Capture(on_prepared)

    def do_join():
      self._zk.acreate(posixpath.join(self._path, self.MEMBER_PREFIX),
          blob, self._acl, zookeeper.SEQUENCE | zookeeper.EPHEMERAL, acreate_completion)

    def exists_watch(_, event, state, path):
      if (event == zookeeper.SESSION_EVENT and state == zookeeper.EXPIRED_SESSION_STATE) or (
          event == zookeeper.DELETED_EVENT):
        exists_capture.set()

    def exists_completion(path, _, rc, stat):
      if rc in self._zk.COMPLETION_RETRY:
        self._zk.aexists(path, exists_watch, functools.partial(exists_completion, path))
        return
      if rc == zookeeper.NONODE:
        exists_capture.set()

    def acreate_completion(_, rc, path):
      if rc in self._zk.COMPLETION_RETRY:
        do_join()
        return
      if rc == zookeeper.OK:
        created_id = self.znode_to_id(path)
        membership = Membership(created_id)
        with self._member_lock:
          result_future = self._members.get(membership, Future())
          result_future.set_result(blob)
          self._members[membership] = result_future
        if expire_callback:
          self._zk.aexists(path, exists_watch, functools.partial(exists_completion, path))
      else:
        membership = Membership.error()
      membership_capture.set(membership)

    self._prepare_path(prepare_capture)
    return membership_capture()

  def cancel(self, member, callback=None):
    capture = Capture(callback)

    def do_cancel():
      self._zk.adelete(posixpath.join(self._path, self.id_to_znode(member.id)),
        -1, adelete_completion)

    def adelete_completion(_, rc):
      if rc in self._zk.COMPLETION_RETRY:
        do_cancel()
        return
      # The rationale here is two situations:
      #   - rc == zookeeper.OK ==> we successfully deleted the znode and the membership is dead.
      #   - rc == zookeeper.NONODE ==> the membership is dead, though we may not have actually
      #      been the ones to cancel, or the node never existed in the first place.  it's possible
      #      we owned the membership but it got severed due to session expiration.
      if rc == zookeeper.OK or rc == zookeeper.NONODE:
        future = self._members.pop(member.id, Future())
        future.set_result(Membership.error())
        capture.set(True)
      else:
        capture.set(False)

    do_cancel()
    return capture()

  def monitor(self, membership=frozenset(), callback=None):
    capture = Capture(callback)

    def wait_exists():
      self._zk.aexists(self._path, exists_watch, exists_completion)

    def exists_watch(_, event, state, path):
      if event == zookeeper.SESSION_EVENT and state == zookeeper.EXPIRED_SESSION_STATE:
        wait_exists()
        return
      if event == zookeeper.CREATED_EVENT:
        do_monitor()
      elif event == zookeeper.DELETED_EVENT:
        wait_exists()

    def exists_completion(_, rc, stat):
      if rc == zookeeper.OK:
        do_monitor()

    def do_monitor():
      self._zk.aget_children(self._path, get_watch, get_completion)

    def get_watch(_, event, state, path):
      if event == zookeeper.SESSION_EVENT and state == zookeeper.EXPIRED_SESSION_STATE:
        wait_exists()
        return
      if state != zookeeper.CONNECTED_STATE:
        return
      if event == zookeeper.DELETED_EVENT:
        wait_exists()
        return
      if event != zookeeper.CHILD_EVENT:
        return
      if set_different(capture, membership, self._members):
        return
      do_monitor()

    def get_completion(_, rc, children):
      if rc in self._zk.COMPLETION_RETRY:
        do_monitor()
        return
      if rc == zookeeper.NONODE:
        wait_exists()
        return
      if rc != zookeeper.OK:
        log.warning('Unexpected get_completion return code: %s' % ReturnCode(rc))
        capture.set(set([Membership.error()]))
        return
      self._update_children(children)
      set_different(capture, membership, self._members)

    do_monitor()
    return capture()

  def list(self):
    try:
      return sorted(map(lambda znode: Membership(self.znode_to_id(znode)),
          filter(self.znode_owned, self._zk.get_children(self._path))))
    except zookeeper.NoNodeException:
      return []


class ActiveGroup(Group):
  """
    An implementation of GroupInterface against CZookeeper when iter() and
    monitor() are expected to be called frequently.  Constantly monitors
    group membership and the contents of group blobs.
  """

  def __init__(self, *args, **kwargs):
    super(ActiveGroup, self).__init__(*args, **kwargs)
    self._monitor_queue = []
    self._monitor_members()

  def monitor(self, membership=frozenset(), callback=None):
    capture = Capture(callback)
    if not set_different(capture, membership, self._members):
      self._monitor_queue.append((membership, capture))
    return capture()

  # ---- private api

  def _monitor_members(self):
    def wait_exists():
     self._zk.aexists(self._path, exists_watch, exists_completion)

    def exists_watch(_, event, state, path):
      if event == zookeeper.SESSION_EVENT and state == zookeeper.EXPIRED_SESSION_STATE:
        wait_exists()
        return
      if event == zookeeper.CREATED_EVENT:
        do_monitor()
      elif event == zookeeper.DELETED_EVENT:
        wait_exists()

    def exists_completion(_, rc, stat):
      if rc == zookeeper.OK:
        do_monitor()

    def do_monitor():
      self._zk.aget_children(self._path, membership_watch, membership_completion)

    def membership_watch(_, event, state, path):
      # Connecting state is caused by transient connection loss, ignore
      if state == zookeeper.CONNECTING_STATE:
        return
      if event == zookeeper.DELETED_EVENT:
        wait_exists()
        return
      # Everything else indicates underlying change.
      do_monitor()

    def membership_completion(_, rc, children):
      if rc in self._zk.COMPLETION_RETRY:
        do_monitor()
        return
      if rc == zookeeper.NONODE:
        wait_exists()
        return
      if rc != zookeeper.OK:
        return

      children = [child for child in children if self.znode_owned(child)]
      _, new = self._update_children(children)
      for child in new:
        def devnull(*args, **kw): pass
        self.info(child, callback=devnull)

      monitor_queue = self._monitor_queue[:]
      self._monitor_queue = []
      members = set(Membership(self.znode_to_id(child)) for child in children)
      for membership, capture in monitor_queue:
        if set(membership) != members:
          capture.set(members)
        else:
          self._monitor_queue.append((membership, capture))

    do_monitor()



from abc import ABCMeta, abstractproperty
from twitter.common.lang import Compatibility


class NamedValue(object):
  __metaclass__ = ABCMeta

  def __init__(self, value):
    if isinstance(value, int):
      self._value = value if value in self.map else 0
    elif isinstance(value, Compatibility.string):
      self._value = dict((v, k) for (k, v) in self.map.items()).get(value.upper(), 0)
    else:
      raise ValueError('Unknown value: %s' % value)

  @abstractproperty
  def map(self):
    """Returns the map from id => string"""
    pass

  def __str__(self):
    return self.map.get(self._value, 'UNKNOWN')

  def __repr__(self):
    return '%s(%r)' % (self.__class__.__name__, self.map[self._value])




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(BuildFileAddress(FilesystemBuildFile(/Users/jdorlus/dev/GitHub/twitter_commons/src/python/twitter/common/zookeeper/BUILD), zookeeper))

from setuptools import setup

setup(**
{   'description': "Implementations of Twitter's service discovery libraries on top of Kazoo.",
    'extras_require': {   'old': ['zc-zookeeper-static==3.4.4']},
    'install_requires': [   'kazoo==2.2.1',
                            'twitter.common.metrics==0.3.9',
                            'twitter.common.concurrent==0.3.9',
                            'twitter.common.lang==0.3.9',
                            'twitter.common.service-thrift==0.3.9',
                            'thrift',
                            'twitter.common.log==0.3.9',
                            'twitter.common.exceptions==0.3.9',
                            'twitter.common.quantity==0.3.9'],
    'name': 'twitter.common.zookeeper',
    'namespace_packages': ['twitter', 'twitter.common'],
    'package_data': {   },
    'package_dir': {   '': 'src'},
    'packages': [   'twitter',
                    'twitter.common',
                    'twitter.common.zookeeper',
                    'twitter.common.zookeeper.group',
                    'twitter.common.zookeeper.serverset'],
    'version': '0.3.9'}
)
